from .cache import Cache
from .cache_util import from_cache_or_fetch
from .cache_util import cache
from .parser import parse_metadata
from .convertor import convert
from .parser import KeyExtract
from .params import build_params
from .http import http_get
from .get_station import get_station



__all__ = [
    "Cache",
    "KeyExtract",
    "from_cache_or_fetch",
    "parse_metadata",
    "convert",
    "build_params",
    "cache",
    "http_get",
    "get_station"
]


from .Route import Route
from .BaseModel import BaseModel

class Vehicle(BaseModel):
    def __init__(
        self,
        route: Route = None, 
        routeId: str = None,
        routeNo: str = None,
        gpsLati: float = None,
        gpsLong: float = None,
        arrtime: int = None,
        arrprevstationcnt: int = None,
        vehicleTp: str = None,
        vehicleNo: str = None
    ):
        self.route = route
        self.routeId = routeId
        self.routeNo = routeNo
        self.gpsLati = gpsLati
        self.gpsLong = gpsLong
        self.arrtime = arrtime
        self.arrprevstationcnt = arrprevstationcnt
        self.vehicleTp = vehicleTp
        self.vehicleNo = vehicleNo
    
    def __repr__(self):
        return f"Vehicle({self.routeNo} - {self.vehicleNo})"
    
    def to_dict(self):
        return vars(self)
    
    @classmethod
    def from_dict(cls, data: dict) -> "Vehicle":
        return cls(
            route=data.get("route"),
            routeId=data.get("routeid"),
            routeNo=data.get("routeno", data.get("routenm")),
            gpsLati=data.get("gpslati"),
            gpsLong=data.get("gpslong"),
            arrtime=data.get("arrtime"),
            arrprevstationcnt=data.get("arrprevstationcnt"),
            vehicleTp=data.get("vehicletp"),
            vehicleNo=data.get("vehicleno")
        )
    
    @classmethod
    def from_list(cls, data:list[dict]) -> list["Vehicle"]:
        return [ cls.from_dict(vehicle) for vehicle in data]




class TAGOAuth:
    def __init__(self, serviceKey: str):
        if not isinstance(serviceKey, str):
            raise TypeError("Service key must be a string")
        
        self._serviceKey = serviceKey

    @property
    def serviceKey(self):
        return self._serviceKey


import pandas as pd
from tagoapi.models import Station
from .cache import Cache

# def csv_to_dict(csvfile, encoding):
#     loaded_csv = pd.read_csv(csvfile, encoding=encoding)
#     data = loaded_csv.to_dict(orient='records')
#     return data
cache = Cache("caches/station.pkl")
station_list = cache.get("stations_2025_06_15.csv")

def get_station(keyword) -> list[Station]:
    result = []
    for station in station_list:
        if  keyword in station["정류장명"]:
            result.append(
                Station.from_dict({
                    "nodeid": station["정류장번호"],
                    "nodenm": station["정류장명"],
                    "nodeno": station["모바일단축번호"],
                    "gpslati": station["위도"],
                    "gpslong": station["경도"],
                    "citycode": station["도시코드"]
                })
            )
    

    return station_list


# if not station_list:
#     station_list = csv_to_dict("tagoapi/csv/stations_2025_06_15.csv", 'cp949')
#     cache.save("stations_2025_06_15.csv", station_list)




from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tagoapi import TAGOAuth

def build_params(
        auth: "TAGOAuth",  
        numOfRows: int = 300, 
        pageNo: int = 1,
        **kwargs: dict
    ) -> dict:
    
    return {
            "serviceKey": auth.serviceKey,
            "numOfRows": numOfRows,
            "pageNo": pageNo,
            "_type": "json",
            **{key: value for key, value in kwargs.items() if value}
        }


from .BaseModel import BaseModel


class ArrivalInfo(BaseModel):
    def __init__(self,
        nodeid,
        nodenm,
        routeId: str,
        routeNo: str,
        routetp,
        arrprevstationcnt: int = None,
        vehicletp: str = None,
        arrtime: int = None
    ):
        self.nodeid = nodeid
        self.nodenm = nodenm
        self.routeId = routeId
        self.routeNo = routeNo
        self.routetp = routetp
        self.arrprevstationcnt = arrprevstationcnt
        self.vehicletp = vehicletp
        self.arrtime = arrtime
    
    def __repr__(self):
        return f"ArrivalInfo({self.routeNo})"
    
    def to_dict(self):
        return vars(self)
    
    @classmethod
    def from_dict(cls, data: dict) -> "ArrivalInfo":
        return cls(
            nodeid=data.get("nodeid"),
            nodenm=data.get("nodeid"),
            routetp=data.get("routetp"),
            routeId = data.get("routeid"),
            routeNo = data.get("routeno"),
            arrprevstationcnt = data.get("arrprevstationcnt"),
            vehicletp = data.get("vehicletp"),
            arrtime = data.get("arrtime")
        )
    
    @classmethod
    def from_list(cls, data: list[dict]) -> list["ArrivalInfo"]:
        return [cls.from_dict(station) for station in data]



from typing import TYPE_CHECKING
from .BaseModel import BaseModel
if TYPE_CHECKING:
    from .Route import Route

class Station(BaseModel):
    # cache_key = "Station:<nodeId><nodenm>"
    cache_key = "Station:<nodeId>"
    
    def __init__(
        self,
        nodeId: str,
        nodeNm: str,
        nodeNo: int = None,
        gpsLati: float = None,
        gpsLong: float = None,
        cityCode: int = None,
        updowncd: int = None,
        nodeord: int = None
        # *routeList: list['Route']
    ):
        self.nodeId = nodeId
        self.nodeNm = nodeNm
        self.nodeNo = nodeNo
        self.gpsLati = gpsLati
        self.gpsLong = gpsLong
        self.cityCode = cityCode
        self.updowncd = updowncd
        self.nodeord = nodeord
        # self.routeList = routeList

    def __repr__(self):
        return f"Station({self.nodeNm})"
    
    def to_dict(self):
        return vars(self)
    
    @classmethod
    def from_dict(cls, data: dict) -> "Station":
        return cls(
            nodeId = data.get("nodeid"),
            nodeNm = data.get("nodenm"),
            nodeNo = data.get("nodeno"),
            gpsLati = float(data.get("gpslati")),
            gpsLong = float(data.get("gpslong")),
            cityCode = data.get("citycode"),
            updowncd = data.get("updowncd"),
            nodeord = data.get("nodeord")
        )
    
    @classmethod
    def from_list(cls, data: list[dict]) -> list["Station"]:
        return [cls.from_dict(station) for station in data]
    


from .Route import Route
from .Vehicle import Vehicle
from .Station import Station
from .ArrivalInfo import ArrivalInfo
from .BaseModel import BaseModel

__all__ = ['Route', 'Vehicle', 'Station', 'ArrivalInfo', 'BaseModel']




class BaseModel:
    cache_key = "BaseModel:<id>"
    def __init__(self): ...

    def to_dict(self) -> dict: ...

    def to_dict(self):
        return vars(self)

    @classmethod
    def from_dict(cls, data:dict) -> "BaseModel": ...

    @classmethod
    def from_list(cls, data:list) -> list["BaseModel"]:
        return [cls.from_dict(v) for v in data]
    





import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from tagoapi import *
from tagoapi.utils.cache import Cache



def test_cache():
    cache = Cache("caches/station.pkl")
    station_list = cache.get("stations_2025_06_15.csv")
    assert isinstance(station_list, list)

def test_statin():
    assert isinstance(get_station("민들"), list)




import os, pickle
from time import time

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
PROJECT_ROOT = os.path.abspath(os.path.join(MODULE_DIR, "..", ".."))
CACHE_DIR = os.path.join(PROJECT_ROOT, "caches")
DEFAULT_CACHE_PATH = os.path.join(CACHE_DIR, "cache.pkl")

class Cache:
    def __init__(self, path: str = DEFAULT_CACHE_PATH):
        self.path = os.path.abspath(path)
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        self._cache = self._load()

    def _load(self) -> dict:
        if os.path.exists(self.path):
            with open(self.path, 'rb') as f:
                return pickle.load(f)
        return {}

    def save(self, key: str, value: dict, ttl: int = 86400) -> bool:
        self._cache[key] = {
            "value": value,
            "ttl": ttl,
            "saved_time": time()
        }
        with open(self.path, 'wb') as f:
            pickle.dump(self._cache, f)
        return True

    def get(self, key: str) -> dict | None:
        entry = self._cache.get(key)
        if not entry:
            return None

        value, saved_time, ttl = entry["value"], entry["saved_time"], entry["ttl"]
        if time() - saved_time > ttl:
            del self._cache[key]
            return None
        return value

    @property
    def current_cache(self):
        return self._cache



from .exceptions import *
from .utils import *
from .models import *
from .auth import TAGOAuth

from typing import Union, Optional, overload



class TAGOClient:
    BASE_URL = "http://apis.data.go.kr/1613000"
    AVRINFO = "ArvlInfoInqireService"
    BUSROUTE = 'BusRouteInfoInqireService'
    BUSTATION = "BusSttnInfoInqireService"
    BUSPOS = "BusLcInfoInqireService"
    CACHE_TTL = 604800


    def __init__(self, auth: TAGOAuth):
        if not isinstance(auth, TAGOAuth):
            raise TypeError("Expected 'auth' to be an instance of TAGOAuth")
        
        self.auth = auth

    @overload
    def get_station(self, cityCode: int, nodeNo: int) -> list[Station]: ...
    @overload
    def get_station(self, cityCode: int, nodeNo: Optional[int], nodeNm: str) -> list[Station]: ...

    @from_cache_or_fetch(604800)
    def get_route_by_no(
        self,
        cityCode: int,
        routeNo: str
    ) -> list[Route]:
        endpoint = f'{self.BUSROUTE}/getRouteNoList'
        params = build_params(self.auth, cityCode=cityCode, routeNo=routeNo)
        return self._fetch_and_convert(endpoint, params, Route)
        
    @from_cache_or_fetch(604800)
    def get_route_by_id(
        self,
        cityCode: int,
        routeId: str
    ) -> Route:
        endpoint = f'{self.BUSROUTE}/getRouteInfoIem'
        params = build_params(self.auth, cityCode=cityCode, routeId=routeId)
        return self._fetch_and_convert(endpoint, params, Route, is_list=False)

    @from_cache_or_fetch()
    def get_route_by_station(
        self,
        cityCode: int,
        nodeId: str
    ) -> list[Route]:
        endpoint = f'{self.BUSTATION}/getSttnThrghRouteList'
        params = build_params(self.auth, cityCode=cityCode, nodeId=nodeId)
        return self._fetch_and_convert(endpoint, params, Route)
    

    @from_cache_or_fetch(604800)
    def get_station_by_route(
        self,
        cityCode: int,
        routeId: str
    ) -> list[Station]:
        endpoint = f'{self.BUSROUTE}/getRouteAcctoThrghSttnList'
        params= build_params(self.auth, cityCode=cityCode, routeId=routeId)
        return self._fetch_and_convert(endpoint, params, Station)
    
    @from_cache_or_fetch(86400)
    def get_station(
        self,
        cityCode: int,
        nodeNo: int = None,
        nodeNm: str = None,
    ) -> list[Station]:
        if not (nodeNo or nodeNm):
            raise ValueError("Only one of 'nodeNo' or 'nodeNm' should be provided.")

        endpoint = f'{self.BUSTATION}/getSttnNoList'
        params= build_params(self.auth, cityCode=cityCode, nodeNm=nodeNm,nodeNo=nodeNo)
        return self._fetch_and_convert(endpoint, params, Station)
    
    @from_cache_or_fetch(86400)
    def get_station_by_gps(
        self,
        gpsLati: float,
        gpsLong: float
    ) -> list[Station]:
        endpoint = f'{self.BUSTATION}/getCrdntPrxmtSttnList'
        params = build_params(self.auth, gpsLati=gpsLati, gpsLong=gpsLong)
        return self._fetch_and_convert(endpoint, params, Station)


    def get_arrival_by_station(
        self,
        cityCode: int,
        nodeId: str,
    ) -> list[ArrivalInfo]:
        endpoint = f'{self.AVRINFO}/getSttnAcctoArvlPrearngeInfoList'
        params = build_params(self.auth, cityCode=cityCode, nodeId=nodeId)
        return self._fetch_and_convert(endpoint, params, ArrivalInfo)
    
    def get_route_arrival_by_station(
        self,
        cityCode: int,
        nodeId: str,
        routeId: str,
    ) -> list[ArrivalInfo]:
        endpoint = f'{self.AVRINFO}/getSttnAcctoSpcifyRouteBusArvlPrearngeInfoList'
        params = build_params(self.auth, cityCode=cityCode, nodeId=nodeId, routeId=routeId)
        return self._fetch_and_convert(endpoint, params, ArrivalInfo)
    

    def get_route_pos(
        self, 
        cityCode: int,
        routeId: int,
    ) -> list[Vehicle]:
        endpoint = f'{self.BUSPOS}/getRouteAcctoBusLcList'
        params = build_params(self.auth, cityCode=cityCode, routeId=routeId)
        return self._fetch_and_convert(endpoint, params, Vehicle, is_cache=False)

    def get_route_pos_near_station(
        self, 
        cityCode: int,
        routeId: int,
        nodeId: int,
    ) -> list[Vehicle]:
        endpoint = f'{self.BUSPOS}/getRouteAcctoSpcifySttnAccesBusLcInfo'
        params = build_params(self.auth, cityCode=cityCode, routeId=routeId, nodeId=nodeId)
        return self._fetch_and_convert(endpoint, params, Vehicle)



    def _fetch_and_convert(
            self, 
            endpoint: str, 
            params: dict, 
            model: BaseModel,
            is_list: bool = True,
            is_cache: bool = True
    ) -> BaseModel:
        cache_key = KeyExtract(model)
        response = parse_metadata(self._get(endpoint, params))
        if not response: 
            return None 
        if isinstance(response, list):
            if not is_cache:
                return convert(response, model.from_list)
            result = []
            for v in response:
                key = cache_key.generate_key(v)
                cached = cache.get(key)
                if cached:
                    result.append(cached)
                else:
                    parsed_obj = convert(v, model.from_dict)
                    result.append(parsed_obj)
                    cache.save(key, parsed_obj, self.CACHE_TTL)

            return result

        else:
            if is_cache: 
                key = cache_key.generate_key(response)
                cached = cache.get(key)
                if cached:
                    return [cached] if is_list else cached
            result = convert(response, model.from_dict)
            cache.save(key, result, self.CACHE_TTL)
            return [result] if is_list else result

    
    def _get(self, endpoint: str, params: dict) -> any:
        response = http_get(f"{self.BASE_URL}/{endpoint}", params=params)
        error_code = response.get("returnReasonCode")
        if error_code == '30':
            raise ServiceKeyNotRegisteredError("유효하지 않는 서비스키 입니다.")
        elif error_code:
            raise RuntimeError(f"실행중 오류가 발생했습니다. 에러코드: {error_code}")

        return response


import requests
import xmltodict
from requests.exceptions import (
    ConnectionError, Timeout, HTTPError, RequestException, JSONDecodeError
)


def http_get(endpoint: str, params: dict) -> dict:
    try:
        response = requests.get(endpoint, params=params, timeout=(3, 10))
        response.raise_for_status()

        return response.json()
    except ConnectionError:
        raise RuntimeError("서버에 연결할 수 없습니다. 인터넷 연결이나 도메인을 확인해주세요.")
    except Timeout:
        raise RuntimeError("요청 시간이 초가되었습니다.")
    except HTTPError as e:
        raise RuntimeError(f"HTTP 오류 발생: {e.response.status_code}")
    except JSONDecodeError as e:
        try:
            return xmltodict.parse(response.text).get("OpenAPI_ServiceResponse", {}).get("cmmMsgHeader", {})
        except Exception as e:
            raise ValueError("응답을 JSON으로 디코딩 할 수 없습니다.")
    

    except RequestException as e:
        raise RuntimeError(f"요청 중 알 수 없는 오류 발생: {e}")


from typing import TYPE_CHECKING
from .BaseModel import BaseModel

if TYPE_CHECKING:
    from .Station import Station


# 버스 노선 자체에 관한 정보
class Route(BaseModel):
    cache_key = "Route:<routeId>"

    def __init__(
        self,
        routeId: str,
        routeNo: str = None,
        routeTp: str = None,
        endNodeNm: str = None,
        startNodeNm: str = None,
        endvehicletime: int = None,
        startvehicletime: int = None,
        #TODO: 정류장 리스트도 넣으면 좋을 듯 합니당
    ):

        self.routeId = routeId
        self.routeNo = routeNo
        self.routeTp = routeTp
        self.endNodeNm = endNodeNm # 다른 곳에서 표시할 땐 이름으로
        self.startNodeNm = startNodeNm
        self.endvehicletime = endvehicletime
        self.startvehicletime = startvehicletime

    def __repr__(self):
        return f"Route({self.routeNo})"
    
    def to_dict(self):
        return vars(self)
    
    @classmethod
    def from_dict(cls, data: dict) -> "Route":
        return cls(
            routeId=data.get("routeid"),
            routeNo=data.get("routeno"),
            routeTp=data.get("routetp"),
            startNodeNm=data.get("startnodenm"),
            endNodeNm=data.get("endnodenm"),
            endvehicletime=data.get("endvehicletime"),
            startvehicletime=data.get("startvehicletime")
        )
    
    @classmethod
    def from_list(cls, data: list[dict]) -> list["Route"]:
        return [ cls.from_dict(route) for route in data ]



from tagoapi.models import BaseModel
from typing import Union, Callable, TypeVar, Any
import os

U = TypeVar('U', dict[str], list) # list, dict으로 반환할 때

def parse_metadata(res: dict) -> U:
    striped = res.get("response", {}).get("body", {}).get("items", {})
    if isinstance(striped, dict):
        return striped.get("item", None)

    return None

class KeyExtract:
    def __init__(self, model: BaseModel):
        self.model = model
        self.raw_key = model.cache_key

        self._args = self._check_bracket(self.raw_key)

    @property
    def key_args(self):
        return self._args
    
    def generate_key(self, data: dict) -> str:
        generated_key = self.raw_key
        # if len(kwargs) > len(self._args):
        #     raise TypeError(f"generate_key() takes {len(self._args)} positional argument but {len(kwargs)} were given")
        
        #TODO: 이거 self._args말고 kwargs.keys해서 arg랑 대응 시켜서 없으면 raise 이런식으로 작성해도 좋을 듯
        for arg in self._args:
            k = data.get(arg.lower(), None)
            if not k:
                raise TypeError()
            generated_key = generated_key.replace(f"<{arg}>", k)

        return generated_key
    
    def __call__(self):
        pass
    
    def _check_bracket(self, data: str) -> list[str]:
        args = []
        current_match = None
        fa = ''
        for w in data:
            if current_match:
                fa += w

            if w == '<':
                current_match = True
            elif w == '>':
                if not current_match:
                    return None # 에러를 나타내든 뭐를 하든 
                current_match = False
                args.append(fa[:-1].strip()); fa = ''

        return args



from setuptools import setup, find_packages

setup(
    name="UnOffical_TAGO_API",
    version="0.2",
    description="test1",
    url="https://github.com/hyuntroll/TAGOBus-API",
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        "requests",
        "xmltodict"
    ],
    license='MIT',
    author='hyuntroll',
    author_email="hsm200905292@gmail.com"
)


# 패키지를 불러올 때 보여줄 것들만 표시
from .client import TAGOClient
from .auth import TAGOAuth


from .models import Route
from .models import Vehicle
from .models import Station

# from .utils import get_city_code
from .utils.cache_util import from_cache_or_fetch
from .utils.parser import KeyExtract
from .utils.cache_util import cache
from .utils.get_station import get_station



__all__ = [ 'TAGOClient', 'TAGOAuth', 'from_cache_or_fetch', 'Route', 'Vehicle', 'Station', 'KeyExtract', 'cache', 'get_station' ]


from .cache import Cache


cache = Cache()   

# method에서만 사용할 함수
def from_cache_or_fetch(ttl: int = 86400): # 데코레이터가 사용할 매개변수
    def real_deco(fn): # 호출할 함수를 매개변수로 받음
        def wrapper(self, *args, **kwargs): # 호출할 함수의 매개변수를 받아서 이를 실행
            key = generate_cache_key(*args, _fname=fn.__name__, **kwargs)
            cached = cache.get(key)
            if cached:
                return cached
            
            result = fn(self, *args, **kwargs)
            cache.save(key, result, ttl)
            return result

        return wrapper
    return real_deco

def generate_cache_key(*args, _fname: str, **kwargs) -> str:
    return _fname + ":" + "&".join([str(a) for a in args]) + "&".join(
        f"{key}={value}" for key, value in kwargs.items()
    ) ## str로 나타낼 수 없으면 다르게 표시하도록


# def get_city_code(serviceKey) -> dict:
#     # 요청 후에 캐시로 저장하는 코드
#     key = "BusSttnInfoInqireService/getCtyCodeList"
#     cached = cache.get(key)
#     if cached:
#         print("cache hit!")
#         return cached
    
#     print("cache miss!")
#     path = "http://apis.data.go.kr/1613000/BusSttnInfoInqireService/getCtyCodeList"
#     params = {'_type': 'json', "serviceKey": serviceKey}
#     res = requests.get(path, params=params).json()

#     cache.save(key, res, 15552000)
#     return res




class TAGOAPIError(Exception):
    """TAGO API 사용중 생기는 일반적인"""
    pass

class UtilError(Exception):
    """Utils 사용 중 생긴 오류"""
    pass
class ServiceKeyNotRegisteredError(TAGOAPIError):
    """등록되지 않은 서비스키를 사용"""
    pass

class RequestExcessdsError(TAGOAPIError):
    """서비스 요청제한횟수 초과 에러"""
    pass

class NoOpenAPIServiceError(TAGOAPIError):
    """오픈 API 서비스가 없거나 페기됨"""
    pass

class TAGORequestError(TAGOAPIError):
    """요청 실패 (http 에러, 타임아웃 등)"""
    def __init__(self, message: str, status_code: int | None ):
        super().__init__(message)
        self.status_code = status_code

class TAGOResponseError(TAGOAPIError):
    """ 요청 오류 """
    def __init__(self, message: str, status_code: int | None ):
        super().__init__(message)
        self.status_code = status_code



from typing import Union, Callable, TypeVar, Any
T = TypeVar('T', dict, list) # list, dict으로 매개변수 받을 때
U = TypeVar('U', dict[str], list) # list, dict으로 반환할 때

def convert(res: T, converter: Callable[[T], U]) -> U:
    if not res:
        return None
    return converter(res)
